
#ifndef _UESQL_BASE_MANAGER__
#define _UESQL_BASE_MANAGER__

#include <algorithm>
#include <mpcl/general-exceptions.hh>
#include <mpcl/string.hh>
#ifdef _UESQL_DEBUG__
#include <mpcl/log_facilities.hh>
#endif  // not _UESQL_DEBUG__
#include <queue>
#include <uesqlc/defs-uesqlc.hh>
#include <vector>


///  Name-space for UESQL.
namespace uesql
{

  ///  Unary function object to delete parameters.
  template <typename TItem>
  struct MUpdateFromBuffer : public std::unary_function<TItem, void>
  {

      void operator () (TItem* ptITEM) const
      {
        ptITEM->updateFromBuffer();
      }

  };  // struct MUpdateFromBuffer

  ///  Unary function object to delete parameters.
  template <typename TItem>
  struct MWriteDebugInfo : public std::unary_function<TItem, void>
  {

      void operator () (TItem* ptITEM) const
      {
#       ifdef _UESQL_DEBUG__
        mpcl::log << ", " << ptITEM->get();
#       endif  // not _UESQL_DEBUG__
      }

  };  // struct MWriteDebugInfo

  ///  UESQL base manager class.
  template <typename TConnection, typename TResult>
  class TUesqlBaseManager : protected TSqlstateDescriptor
  {
  
    protected:

      ///  Map type that binds connection names to connections.
      typedef
        mpcl::xextended_map<mpcl::xstring, TConnection>
        TConnectionMap;

      ///  Boolean with truth value only when there is an active connection.
      bool   gIsThereConnection;
  
      ///  Integer with the ordinal of the next unique connection name.
      mutable int   iUniqueOrdinal;
  
      ///  Map that binds connection names to connections.
      TConnectionMap   tConnectionMap;

      ///  Current connection (from \a tConnectionMap).
      TConnection   tCurrentConnection;

      ///  Input embedded variable string values queue.
      std::queue<mpcl::xstring>   tInputEmbeddedVariableQueue;

      ///  Output embedded variable string pointers vector.
      std::vector<TSqlDataType*>   tOutputEmbeddedVariableVector;

      ///  String with the identifier of the current connection.
      mpcl::xstring   yCurrentConnection;


    public:
  
      //
      //  C O N S T R U C T O R S
      //
  
      ///  Builds a new instance.
      TUesqlBaseManager (void)                :
        TSqlstateDescriptor           ()      ,
        gIsThereConnection            (false) ,
        iUniqueOrdinal                (0)     ,
        tConnectionMap                ()      ,
        tCurrentConnection            ()      ,
        tInputEmbeddedVariableQueue   ()      ,
        tOutputEmbeddedVariableVector ()      ,
        yCurrentConnection            ()
      {
        initialize();
      }

      ///  Destroys the instance.
      virtual ~TUesqlBaseManager (void)
      {
        finish();
      }

      /**
      *  Binds in connection map, connection name \a rkyCONNECTION_NAME to
      *  \a rtCONNECTION.
      *  @param rkyCONNECTION_NAME Connection name.
      *  @param rtCONNECTION       Connection.
      */
      virtual void connect ( const mpcl::xstring& rkyCONNECTION_NAME ,
                             TConnection&         rtCONNECTION       );

      /**
      *  Diagnoses statement  result  (\a tRESULT). It only cheks the result of
      *  the statement executed.  This member function (\a diagnose())  updates
      *  properly SQL status variables.
      *  @param tRESULT Source statement execution result.
      */
      virtual void diagnose (TResult tRESULT) = 0;

      /**
      *  Disconnects connection named \a rkyCONNECTION_NAME.
      *  @param rkyCONNECTION_NAME Source connection name.
      *  @warning kyCONNECTION_NAME  mustn't  be  passed  as reference to avoid
      *           problems  about passing  \a yCurrentConnection  as  argument.
      */
      virtual void disconnect (const mpcl::xstring kyCONNECTION_NAME);
  
      ///  Disconnects current connection (if any).
      virtual void disconnectCurrent (void)
      {
        if ( gIsThereConnection )
        {
          disconnect (yCurrentConnection);
        }
      }

      ///  Disconnects all connections.
      virtual void disconnectAll (void);
  
      ///  Clears embedded variable containers.
      virtual void clearQueues (void);

      /**
      *  Clears the result of the last executed statement. The
      *  diagnostic  member  functions,  use  that results  to
      *  diagnose the execution.
      */
      virtual void clearResult (TResult tRESULT) = 0;

      /**
      *  Executes the statement rkySTATEMENT instantiating
      *  the input embedded variables (tagged with '$z').
      *  @param rkySTATEMENT String with the source statement.
      */
      virtual TResult execute (const mpcl::xstring& rkySTATEMENT)
      {
        return execute (rkySTATEMENT.c_str());
      }

      /**
      *  Executes the statement  \a rkySTATEMENT instantiating the input
      *  embedded variables (tagged with '$z') in the \a determinate connection
      *  named \a rkyCONNECTION_NAME.
      *  @param rkySTATEMENT       String with the source statement.
      *  @param rkyCONNECTION_NAME Source connection name.
      */
      virtual TResult execute ( const mpcl::xstring&     rkySTATEMENT       ,
                                const mpcl::xstring& rkyCONNECTION_NAME )
      {
        return execute (rkySTATEMENT.c_str(), rkyCONNECTION_NAME.c_str());
      }

      /**
      *  Executes the statement pkcSTATEMENT instantiating
      *  the input embedded variables (tagged with '$z').
      *  @param pkcSTATEMENT String with the source statement.
      */
      virtual TResult execute (const char* pkcSTATEMENT) = 0;

      /**
      *  Executes the statement  \a pkcSTATEMENT instantiating the input
      *  embedded variables (tagged with '$z') in the \a determinate connection
      *  named \a rkyCONNECTION_NAME.
      *  @param pkcSTATEMENT       String with the source statement.
      *  @param pkcCONNECTION_NAME Source connection name.
      */
      virtual TResult execute ( const char* pkcSTATEMENT       ,
                                const char* pkyCONNECTION_NAME ) = 0;

      /**
      *  Executes the statement pkcSTATEMENT instantiating
      *  the input embedded variables (tagged with '$z'). The result
      *  of the execution is omitted.
      *  @param pkcSTATEMENT String with the source statement.
      */
      virtual void executeClean (const char* pkcSTATEMENT)
      {
        clearResult (execute (pkcSTATEMENT));
      }

      ///  Finishes all managed issues.
      virtual void finish (void)
      {
        disconnectAll();
        clearQueues();
        tConnectionMap.clear();
        yCurrentConnection.erase();
      }

      /**
      *  Gets a pointer to the SQL data type of the output embedded variable at
      *  position zINDEX in the output embedded variable vector.
      *  @param zINDEX Position of the variable.
      *  @return Pointer to the output embedded variable.
      */
      virtual TSqlDataType* getFromOutputQueue (const size_t zINDEX) const;

      /**
      *  Initializes the default connection.
      */
      virtual void initialize (void) {}

      /**
      *  Returns a string that is result of instanting the input
      *  embedded variables (if any).
      *  @param pkcSTATEMENT Source statement.
      *  @return The instantiated string.
      */
      mpcl::xstring instantiate (const char* pkcSTATEMENT);
  
      /**
      *  Returns a string that is result of instanting the input
      *  embedded variables (if any).
      *  @param rkySTATEMENT Source statement.
      *  @return The instantiated version of string \a rkySTATEMENT.
      */
      mpcl::xstring instantiate (const mpcl::xstring& rkySTATEMENT)
      {
        return instantiate (rkySTATEMENT.c_str());
      }

      /**
      *  Pushes on the input embedded variable queue the variable
      *  value \a rktEMBEDDED_VALUE.
      *  @param rktEMBEDDED_VALUE Input embedded variable value.
      */
      virtual void pushOnInputQueue (const TSqlDataType& rktEMBEDDED_VALUE)
      {
        tInputEmbeddedVariableQueue.push (rktEMBEDDED_VALUE.get());
      }

      /**
      *  Pushes on the input embedded variable queue the variable
      *  value \a rktEMBEDDED_VALUE.
      *  @param rktEMBEDDED_VALUE Input embedded variable value.
      */
      virtual void pushOnInputQueue (const TSqlString& rkyEMBEDDED_VALUE)
      {
        tInputEmbeddedVariableQueue.push (rkyEMBEDDED_VALUE);
      }
  
      /**
      *  Pushes on the input embedded variable queue the variable
      *  value \a rktEMBEDDED_VALUE.
      *  @param rktEMBEDDED_VALUE Input embedded variable value.
      */
      virtual void pushOnInputQueue (const mpcl::xstring& rkyEMBEDDED_VALUE)
      {
        tInputEmbeddedVariableQueue.push (rkyEMBEDDED_VALUE);
      }

      /**
      *  Pushes on the output embedded variable vector the variable
      *  reference \a rtSQL_DATA_TYPE.
      *  @param rtSQL_DATA_TYPE Output embedded variable reference.
      */
      virtual void pushOnOutputQueue (TSqlDataType* ptSQL_DATA_TYPE)
      {
        tOutputEmbeddedVariableVector.push_back (ptSQL_DATA_TYPE);
      }

      ///  Resets SQL status variables.
      void resetStatus (void);

      /**
      *  Sets current connection to connection named rkyCONNECTION_NAME.
      *  @param rkyCONNECTION_NAME Source connection name.
      */
      void setCurrentConnection (const mpcl::xstring& rkyCONNECTION_NAME);

      /**
      *  Updates output variables  (which  have  been updated passively using a
      *  temporary buffer).
      */
      virtual void updateOutputVariables (void);

  
    public:
  
      //
      //  S E L E C T O R S
      //
  
      /**
      *  Returns the connection name for the current connection.
      *  @return The connection name for the current connection.
      */
      const mpcl::xstring& currentConnectionName (void) const
      {
        return yCurrentConnection;
      }
  
      /**
      *  Returns true if there is a default connection.
      *  @return True if there is a default connection.
      */
      virtual bool existsDefaultConnection (void) const
      {
        return !yCurrentConnection.empty();
      }
  
      /**
      *  Returns true if there is an active connection.
      *  @param rkyCONNECTION_NAME Source connection name to check.
      *  @return True if there is an active connection.
      */
      bool isConnected (const mpcl::xstring& rkyCONNECTION_NAME) const
      {
        return tConnectionMap.is_bound (rkyCONNECTION_NAME);
      }

      /**
      *  Checks if a NOT FOUND condition has happened.
      *  @return True if it has happened.
      */
      bool isNotFoundCondition (void) const
      {
        return ( SQLSTATE == "02000" );
      }

      /**
      *  Checks if a SQLERROR condition has happened.
      *  @return True if it has happened.
      */
      bool isSqlerrorCondition (void) const
      {
        return ( SQLCODE );
      }

  };  // class TUesqlBaseManager

}  // namespace uesql  


//
//  L O C A L   F U N C T I O N S
//

//
//  C O N S T R U C T O R S
//

template <typename TConnection, typename TResult>
inline void uesql::TUesqlBaseManager<TConnection, TResult>::clearQueues (void)
{

  while ( !tInputEmbeddedVariableQueue.empty() )
  {
    tInputEmbeddedVariableQueue.pop();
  }
  tOutputEmbeddedVariableVector.clear();

}  // clearQueues()


template <typename TConnection, typename TResult>
inline void uesql::TUesqlBaseManager<TConnection, TResult>::
connect ( const mpcl::xstring& rkyCONNECTION_NAME ,
          TConnection&         rtCONNECTION       )
{

  tConnectionMap.bind (rkyCONNECTION_NAME, rtCONNECTION);

}  // connect()


template <typename TConnection, typename TResult>
inline void uesql::TUesqlBaseManager<TConnection, TResult>::
disconnect (const mpcl::xstring kyCONNECTION_NAME)
{

  tConnectionMap.erase (kyCONNECTION_NAME);
  if ( kyCONNECTION_NAME == yCurrentConnection )
  {
    yCurrentConnection.erase();
    gIsThereConnection = false;
  }

}  // disconnect()


template <typename TConnection, typename TResult>
inline void uesql::TUesqlBaseManager<TConnection, TResult>::
disconnectAll (void)
{

  while ( !tConnectionMap.empty() )
  {
    disconnect (tConnectionMap.begin()->first);
  }

}  // disconnectAll()


template <typename TConnection, typename TResult>
inline uesql::TSqlDataType* uesql::TUesqlBaseManager<TConnection, TResult>::
getFromOutputQueue (const size_t zINDEX) const
{

  if ( zINDEX >= tOutputEmbeddedVariableVector.size() )
  {
    throw mpcl::TExceptionEL_Constraint ("out of range", __FILE__, __LINE__);
  }
  return tOutputEmbeddedVariableVector [zINDEX];

}  // getFromOutputQueue()


template <typename TConnection, typename TResult>
inline mpcl::xstring uesql::TUesqlBaseManager<TConnection, TResult>::
instantiate (const char* pkcSTATEMENT)
{

  register mpcl::xstring::size_type   zPatternOffset;
  mpcl::xstring                       yStatementInstance (pkcSTATEMENT);

  while ( !tInputEmbeddedVariableQueue.empty() )
  {
    if ( mpcl::xstring::npos == (zPatternOffset = yStatementInstance.find ("$z")) )
    {
      break;
    }
    else
    {
      yStatementInstance.replace (zPatternOffset, 2, tInputEmbeddedVariableQueue.front());
      tInputEmbeddedVariableQueue.pop();
    }
  }
  return yStatementInstance;

}  // instantiate()


template <typename TConnection, typename TResult>
void uesql::TUesqlBaseManager<TConnection, TResult>::
resetStatus (void)
{

  //
  //  Condition:  successful completion
  // 
  SQLCODE    = 0;
  SQLMESSAGE = "";
  SQLSTATE   = "00000";

}  // resetStatus()


template <typename TConnection, typename TResult>
inline void uesql::TUesqlBaseManager<TConnection, TResult>::
setCurrentConnection (const mpcl::xstring& rkyCONNECTION_NAME)
{

  if ( tConnectionMap.is_bound (rkyCONNECTION_NAME) )
  {
    yCurrentConnection = rkyCONNECTION_NAME;
    gIsThereConnection = true;
    tCurrentConnection = tConnectionMap [rkyCONNECTION_NAME];
#   ifdef _UESQL_DEBUG__
    mpcl::log << mpcl::header << "SET CONNECTION '" << rkyCONNECTION_NAME << "';\n";
#   endif  // not _UESQL_DEBUG__
    resetStatus();
  }
  else
  {
    //
    //  Condition:           connection exception
    //  Class subcondition:  connection does not exist
    //
    SQLCODE    = -1;
    SQLMESSAGE = "";
    SQLSTATE   = "08003";
  }

}  // setCurrentConnection()


template <typename TConnection, typename TResult>
inline void uesql::TUesqlBaseManager<TConnection, TResult>::
updateOutputVariables (void)
{

  std::for_each ( tOutputEmbeddedVariableVector.begin()   ,
                  tOutputEmbeddedVariableVector.end()     ,
                  MUpdateFromBuffer<TSqlDataType>() );
# ifdef _UESQL_DEBUG__
  mpcl::log << mpcl::header << "FETCHED (";
  if ( !tOutputEmbeddedVariableVector.empty() )
  {
    mpcl::log << tOutputEmbeddedVariableVector.front()->get();
    std::for_each ( tOutputEmbeddedVariableVector.begin() + 1 ,
                    tOutputEmbeddedVariableVector.end()       ,
                    MWriteDebugInfo<TSqlDataType>()     );
  }
  mpcl::log << ");\n";
# endif  // not _UESQL_DEBUG__

}  // updateOutputVariables()


#endif  // not _UESQL_BASE_MANAGER__
